# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - dotted paths!
 - per-attr repr transform / filter
 - __ne__ ? cases where it still matters
 - ordering ?
 - repr_filter: ta.Union[ta.Callable[[ta.Any], ta.Optional[str]], ta.Literal['not_none', 'truthy']]] ?
  - unify repr/repr_fn/repr_filter
"""
import types  # noqa
import typing as ta


T = ta.TypeVar('T')


##


@ta.final
class AttrOps(ta.Generic[T]):
    @ta.final
    class Attr:
        def __init__(
                self,
                name: str,
                *,
                display: ta.Optional[str] = None,

                repr: bool = True,  # noqa
                repr_fn: ta.Optional[ta.Callable[[ta.Any], ta.Optional[str]]] = None,

                hash: bool = True,  # noqa
                eq: bool = True,
        ) -> None:
            if '.' in name:
                raise NotImplementedError('Dotted paths not yet supported')
            if not name.isidentifier() or name.startswith('__'):
                raise AttributeError(f'Invalid attr: {name!r}')
            self._name = name

            if display is None:
                display = name[1:] if name.startswith('_') and len(name) > 1 else name
            self._display = display

            self._repr = repr
            self._repr_fn = repr_fn

            self._hash = hash
            self._eq = eq

        @classmethod
        def of(
                cls,
                o: ta.Union[
                    'AttrOps.Attr',
                    str,
                    ta.Tuple[str, ta.Union[str, ta.Mapping[str, ta.Any]]],
                    ta.Mapping[str, ta.Any],
                ],
        ) -> 'AttrOps.Attr':
            if isinstance(o, AttrOps.Attr):
                return o
            elif isinstance(o, str):
                return cls(o)
            elif isinstance(o, tuple):
                name, x = o
                kw: ta.Mapping[str, ta.Any]
                if isinstance(x, str):
                    kw = dict(display=x)
                elif isinstance(x, ta.Mapping):
                    kw = x
                else:
                    raise TypeError(x)
                return cls(name, **kw)
            elif isinstance(o, ta.Mapping):
                return cls(**o)
            else:
                raise TypeError(o)

        @property
        def name(self) -> str:
            return self._name

        @property
        def display(self) -> str:
            return self._display

        @property
        def hash(self) -> bool:
            return self._hash

        @property
        def eq(self) -> bool:
            return self._eq

    @staticmethod
    def opt_repr(o: ta.Any) -> ta.Optional[str]:
        return repr(o) if o is not None else None

    @staticmethod
    def truthy_repr(o: ta.Any) -> ta.Optional[str]:
        return repr(o) if o else None

    #

    @ta.overload
    def __init__(
            self,
            *attrs: ta.Sequence[ta.Union[
                str,
                ta.Tuple[str, ta.Union[str, ta.Mapping[str, ta.Any]]],
                ta.Mapping[str, ta.Any],
                Attr,
            ]],

            with_module: bool = False,
            use_qualname: bool = False,
            with_id: bool = False,
            terse: bool = False,
            repr_filter: ta.Optional[ta.Callable[[ta.Any], bool]] = None,
            recursive: bool = False,

            cache_hash: ta.Union[bool, str] = False,
            subtypes_eq: bool = False,
    ) -> None:
        ...

    @ta.overload
    def __init__(
            self,
            attrs_fn: ta.Callable[[T], ta.Tuple[ta.Union[
                ta.Any,
                ta.Tuple[ta.Any, ta.Union[str, ta.Mapping[str, ta.Any]]],
                Attr,
            ], ...]],
            /,
            *,

            with_module: bool = False,
            use_qualname: bool = False,
            with_id: bool = False,
            terse: bool = False,
            repr_filter: ta.Optional[ta.Callable[[ta.Any], bool]] = None,
            recursive: bool = False,

            cache_hash: ta.Union[bool, str] = False,
            subtypes_eq: bool = False,
    ) -> None:
        ...

    def __init__(
            self,
            *args,

            with_module=False,
            use_qualname=False,
            with_id=False,
            terse=False,
            repr_filter=None,
            recursive=False,

            cache_hash=False,
            subtypes_eq=False,
    ) -> None:
        if args and len(args) == 1 and callable(args[0]):
            self._attrs: ta.Sequence[AttrOps.Attr] = self._capture_attrs(args[0])
        else:
            self._attrs = list(map(AttrOps.Attr.of, args))

        self._with_module: bool = with_module
        self._use_qualname: bool = use_qualname
        self._with_id: bool = with_id
        self._terse: bool = terse
        self._repr_filter: ta.Optional[ta.Callable[[ta.Any], bool]] = repr_filter
        self._recursive: bool = recursive

        self._cache_hash: ta.Union[bool, str] = cache_hash
        self._subtypes_eq: bool = subtypes_eq

    @property
    def attrs(self) -> ta.Sequence[Attr]:
        return self._attrs

    #

    @ta.final
    class _AttrCapturer:
        def __init__(self, fn):
            self.__fn = fn

        def __getattr__(self, attr):
            return self.__fn(self, attr)

    @classmethod
    def _capture_attrs(cls, fn: ta.Callable) -> ta.Sequence[Attr]:
        def access(parent, attr):
            dct[(ret := cls._AttrCapturer(access))] = (parent, attr)
            return ret

        dct: dict = {}
        raw = fn(root := cls._AttrCapturer(access))

        def rec(cap):  # noqa
            if cap is root:
                return
            parent, attr = dct[cap]
            yield from rec(parent)
            yield attr

        attrs: ta.List[AttrOps.Attr] = []
        for o in raw:
            if isinstance(o, (AttrOps.Attr, ta.Mapping)):
                attrs.append(AttrOps.Attr.of(o))
                continue

            kw: ta.Mapping[str, ta.Any]
            if isinstance(o, tuple):
                cap, x = o
                if isinstance(x, str):
                    kw = dict(display=x)
                elif isinstance(x, ta.Mapping):
                    kw = x
                else:
                    raise TypeError(x)
            else:
                cap, kw = o, {}

            path = tuple(rec(cap))

            attrs.append(AttrOps.Attr(
                '.'.join(path),
                **kw,
            ))

        return attrs

    #

    _repr: ta.Callable[[T], str]

    @property
    def repr(self) -> ta.Callable[[T], str]:
        try:
            return self._repr
        except AttributeError:
            pass

        def _repr(o: T) -> str:
            vs: ta.List[str] = []
            for a in self._attrs:
                if not a._repr:  # noqa
                    continue
                v = getattr(o, a._name)  # noqa
                if self._repr_filter is not None and not self._repr_filter(v):
                    continue
                if (rfn := a._repr_fn) is None:  # noqa
                    rfn = repr
                if (vr := rfn(v)) is None:
                    continue
                if self._terse:
                    vs.append(vr)
                else:
                    vs.append(f'{a._display}={vr}')  # noqa

            return (
                f'{o.__class__.__module__ + "." if self._with_module else ""}'
                f'{o.__class__.__qualname__ if self._use_qualname else o.__class__.__name__}'
                f'{("@" + hex(id(o))[2:]) if self._with_id else ""}'  # noqa
                f'({", ".join(vs)})'
            )

        if self._recursive:
            _repr = self._reprlib().recursive_repr()(_repr)

        self._repr = _repr
        return _repr

    _reprlib_: ta.ClassVar[ta.Any]

    @classmethod
    def _reprlib(cls) -> ta.Any:
        try:
            return cls._reprlib_
        except AttributeError:
            pass

        import reprlib  # noqa

        cls._reprlib_ = reprlib
        return reprlib

    #

    _DEFAULT_CACHED_HASH_ATTR: ta.ClassVar[str] = '__cached_hash__'

    _hash: ta.Callable[[T], int]

    @property
    def hash(self) -> ta.Callable[[T], int]:
        try:
            return self._hash
        except AttributeError:
            pass

        def _calc_hash(o: T) -> int:
            return hash(tuple(
                getattr(o, a._name)  # noqa
                for a in self._attrs
                if a._hash  # noqa
            ))

        if (ch := self._cache_hash) is not False:
            if ch is True:
                cha = self._DEFAULT_CACHED_HASH_ATTR
            elif isinstance(ch, str):
                cha = ch
            else:
                raise TypeError(ch)

            def _cached_hash(o: T) -> int:
                try:
                    return object.__getattribute__(o, cha)
                except AttributeError:
                    object.__setattr__(o, cha, h := _calc_hash(o))
                return h

            _hash = _cached_hash

        else:
            _hash = _calc_hash

        self._hash = _hash
        return _hash

    #

    _eq: ta.Callable[[T, ta.Any], ta.Union[bool, 'types.NotImplementedType']]

    @property
    def eq(self) -> ta.Callable[[T, ta.Any], ta.Union[bool, 'types.NotImplementedType']]:
        try:
            return self._eq
        except AttributeError:
            pass

        def _eq(o: T, x: ta.Any) -> 'ta.Union[bool, types.NotImplementedType]':
            if self._subtypes_eq:
                if not isinstance(x, type(o)):
                    return NotImplemented
            else:
                if type(x) is not type(o):
                    return NotImplemented

            return all(
                getattr(o, a._name) == getattr(x, a._name)  # noqa
                for a in self._attrs
                if a._eq  # noqa
            )

        self._eq = _eq
        return _eq

    @property
    def hash_eq(self) -> ta.Tuple[
        ta.Callable[[T], int],
        ta.Callable[[T, ta.Any], ta.Union[bool, 'types.NotImplementedType']],
    ]:
        return (self.hash, self.eq)

    #

    @property
    def repr_hash_eq(self) -> ta.Tuple[
        ta.Callable[[T], str],
        ta.Callable[[T], int],
        ta.Callable[[T, ta.Any], ta.Union[bool, 'types.NotImplementedType']],
    ]:
        return (self.repr, self.hash, self.eq)

    #

    class NOT_SET:  # noqa
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    def install(
            self,
            locals_dct: ta.MutableMapping[str, ta.Any],
            *,
            repr: ta.Union[bool, ta.Type[NOT_SET]] = NOT_SET,  # noqa
            hash: ta.Union[bool, ta.Type[NOT_SET]] = NOT_SET,  # noqa
            eq: ta.Union[bool, ta.Type[NOT_SET]] = NOT_SET,
    ) -> 'AttrOps[T]':
        if all(a is self.NOT_SET for a in (repr, hash, eq)):
            repr = hash = eq = True  # noqa
        if repr:
            locals_dct.update(__repr__=self.repr)
        if hash:
            locals_dct.update(__hash__=self.hash)
        if eq:
            locals_dct.update(__eq__=self.eq)
        return self


attr_ops = AttrOps[ta.Any]


##


def attr_repr(obj: ta.Any, *attrs: str, **kwargs: ta.Any) -> str:
    return AttrOps(*attrs, **kwargs).repr(obj)



import abc
import collections.abc
import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ... import reflect as rfl
from ..base.contexts import UnmarshalContext
from ..base.contexts import UnmarshalFactoryContext
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from ..base.values import Value
from .metadata import FieldTypeTagging
from .metadata import Impls
from .metadata import Polymorphism
from .metadata import TypeTagging
from .metadata import WrapperTypeTagging


##


class PolymorphismUnmarshaler(Unmarshaler, lang.Abstract):
    @abc.abstractmethod
    def get_impls(self) -> ta.Mapping[str, Unmarshaler]:
        raise NotImplementedError


@dc.dataclass(frozen=True)
class WrapperPolymorphismUnmarshaler(PolymorphismUnmarshaler):
    m: ta.Mapping[str, Unmarshaler]

    def get_impls(self) -> ta.Mapping[str, Unmarshaler]:
        return self.m

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        ma = check.isinstance(v, collections.abc.Mapping)
        [(tag, iv)] = ma.items()
        u = self.m[tag]
        return u.unmarshal(ctx, iv)


@dc.dataclass(frozen=True)
class FieldPolymorphismUnmarshaler(PolymorphismUnmarshaler):
    m: ta.Mapping[str, Unmarshaler]
    tf: str

    def get_impls(self) -> ta.Mapping[str, Unmarshaler]:
        return self.m

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        ma = dict(check.isinstance(v, collections.abc.Mapping))
        tag = ma.pop(self.tf)
        u = self.m[tag]
        return u.unmarshal(ctx, ma)


def make_polymorphism_unmarshaler(
        impls: Impls,
        tt: TypeTagging,
        ctx: UnmarshalFactoryContext,
) -> Unmarshaler:
    m = {
        t: u
        for i in impls
        for u in [ctx.make_unmarshaler(i.ty)]
        for t in [i.tag, *i.alts]
    }
    if isinstance(tt, WrapperTypeTagging):
        return WrapperPolymorphismUnmarshaler(m)
    elif isinstance(tt, FieldTypeTagging):
        return FieldPolymorphismUnmarshaler(m, tt.field)
    else:
        raise TypeError(tt)


@dc.dataclass(frozen=True)
class PolymorphismUnmarshalerFactory(UnmarshalerFactory):
    p: Polymorphism
    tt: TypeTagging = WrapperTypeTagging()

    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if rty is not self.p.ty:
            return None
        return lambda: make_polymorphism_unmarshaler(self.p.impls, self.tt, ctx)



# ruff: noqa: UP006 UP007 UP045
"""
TODO:
 - def maybe(v: lang.Maybe[T])
 - def not_ ?
 - ** class @dataclass Raise - user message should be able to be an exception type or instance or factory
"""
import collections
import threading
import typing as ta


T = ta.TypeVar('T')
SizedT = ta.TypeVar('SizedT', bound=ta.Sized)

CheckMessage = ta.Union[str, ta.Callable[..., ta.Optional[str]], None]  # ta.TypeAlias
CheckLateConfigureFn = ta.Callable[['Checks'], None]  # ta.TypeAlias
CheckOnRaiseFn = ta.Callable[[Exception], None]  # ta.TypeAlias
CheckExceptionFactory = ta.Callable[..., Exception]  # ta.TypeAlias
CheckArgsRenderer = ta.Callable[..., ta.Optional[str]]  # ta.TypeAlias


##


class Checks:
    def __init__(self) -> None:
        super().__init__()

        self._config_lock = threading.RLock()
        self._on_raise_fns: ta.Sequence[CheckOnRaiseFn] = []
        self._exception_factory: CheckExceptionFactory = Checks.default_exception_factory
        self._args_renderer: ta.Optional[CheckArgsRenderer] = None
        self._late_configure_fns: ta.Sequence[CheckLateConfigureFn] = []

    @staticmethod
    def default_exception_factory(exc_cls: ta.Type[Exception], *args, **kwargs) -> Exception:
        return exc_cls(*args, **kwargs)  # noqa

    #

    def register_on_raise(self, fn: CheckOnRaiseFn) -> None:
        with self._config_lock:
            self._on_raise_fns = [*self._on_raise_fns, fn]

    def unregister_on_raise(self, fn: CheckOnRaiseFn) -> None:
        with self._config_lock:
            self._on_raise_fns = [e for e in self._on_raise_fns if e != fn]

    #

    def register_on_raise_breakpoint_if_env_var_set(self, key: str) -> None:
        import os

        def on_raise(exc: Exception) -> None:  # noqa
            if key in os.environ:
                breakpoint()  # noqa

        self.register_on_raise(on_raise)

    #

    def set_exception_factory(self, factory: CheckExceptionFactory) -> None:
        self._exception_factory = factory

    def set_args_renderer(self, renderer: ta.Optional[CheckArgsRenderer]) -> None:
        self._args_renderer = renderer

    #

    def register_late_configure(self, fn: CheckLateConfigureFn) -> None:
        with self._config_lock:
            self._late_configure_fns = [*self._late_configure_fns, fn]

    def _late_configure(self) -> None:
        if not self._late_configure_fns:
            return

        with self._config_lock:
            if not (lc := self._late_configure_fns):
                return

            for fn in lc:
                fn(self)

            self._late_configure_fns = []

    #

    class _ArgsKwargs:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    def _raise(
            self,
            exception_type: ta.Type[Exception],
            default_message: str,
            message: CheckMessage,
            ak: _ArgsKwargs = _ArgsKwargs(),
            *,
            render_fmt: ta.Optional[str] = None,
    ) -> ta.NoReturn:
        exc_args = ()
        if callable(message):
            message = ta.cast(ta.Callable, message)(*ak.args, **ak.kwargs)
            if isinstance(message, tuple):
                message, *exc_args = message  # type: ignore

        if message is None:
            message = default_message

        self._late_configure()

        if render_fmt is not None and (af := self._args_renderer) is not None:
            rendered_args = af(render_fmt, *ak.args)
            if rendered_args is not None:
                message = f'{message} : {rendered_args}'

        exc = self._exception_factory(
            exception_type,
            message,
            *exc_args,
            *ak.args,
            **ak.kwargs,
        )

        for fn in self._on_raise_fns:
            fn(exc)

        raise exc

    #

    def _unpack_isinstance_spec(self, spec: ta.Any) -> tuple:
        if isinstance(spec, type):
            return (spec,)
        if not isinstance(spec, tuple):
            spec = (spec,)
        if None in spec:
            spec = tuple(filter(None, spec)) + (None.__class__,)  # noqa
        if ta.Any in spec:
            spec = (object,)
        return spec

    @ta.overload
    def isinstance(self, v: ta.Any, spec: ta.Type[T], msg: CheckMessage = None) -> T:
        ...

    @ta.overload
    def isinstance(self, v: ta.Any, spec: ta.Any, msg: CheckMessage = None) -> ta.Any:
        ...

    def isinstance(self, v, spec, msg=None):
        if not isinstance(v, self._unpack_isinstance_spec(spec)):
            self._raise(
                TypeError,
                'Must be instance',
                msg,
                Checks._ArgsKwargs(v, spec),
                render_fmt='not isinstance(%s, %s)',
            )

        return v

    @ta.overload
    def of_isinstance(self, spec: ta.Type[T], msg: CheckMessage = None) -> ta.Callable[[ta.Any], T]:
        ...

    @ta.overload
    def of_isinstance(self, spec: ta.Any, msg: CheckMessage = None) -> ta.Callable[[ta.Any], ta.Any]:
        ...

    def of_isinstance(self, spec, msg=None):
        def inner(v):
            return self.isinstance(v, self._unpack_isinstance_spec(spec), msg)

        return inner

    def cast(self, v: ta.Any, cls: ta.Type[T], msg: CheckMessage = None) -> T:
        if not isinstance(v, cls):
            self._raise(
                TypeError,
                'Must be instance',
                msg,
                Checks._ArgsKwargs(v, cls),
            )

        return v

    def of_cast(self, cls: ta.Type[T], msg: CheckMessage = None) -> ta.Callable[[T], T]:
        def inner(v):
            return self.cast(v, cls, msg)

        return inner

    def not_isinstance(self, v: T, spec: ta.Any, msg: CheckMessage = None) -> T:  # noqa
        if isinstance(v, self._unpack_isinstance_spec(spec)):
            self._raise(
                TypeError,
                'Must not be instance',
                msg,
                Checks._ArgsKwargs(v, spec),
                render_fmt='isinstance(%s, %s)',
            )

        return v

    def of_not_isinstance(self, spec: ta.Any, msg: CheckMessage = None) -> ta.Callable[[T], T]:
        def inner(v):
            return self.not_isinstance(v, self._unpack_isinstance_spec(spec), msg)

        return inner

    ##

    def issubclass(self, v: ta.Type[T], spec: ta.Any, msg: CheckMessage = None) -> ta.Type[T]:  # noqa
        if not issubclass(v, spec):
            self._raise(
                TypeError,
                'Must be subclass',
                msg,
                Checks._ArgsKwargs(v, spec),
                render_fmt='not issubclass(%s, %s)',
            )

        return v

    def not_issubclass(self, v: ta.Type[T], spec: ta.Any, msg: CheckMessage = None) -> ta.Type[T]:
        if issubclass(v, spec):
            self._raise(
                TypeError,
                'Must not be subclass',
                msg,
                Checks._ArgsKwargs(v, spec),
                render_fmt='issubclass(%s, %s)',
            )

        return v

    #

    def in_(self, v: T, c: ta.Container[T], msg: CheckMessage = None) -> T:
        if v not in c:
            self._raise(
                ValueError,
                'Must be in',
                msg,
                Checks._ArgsKwargs(v, c),
                render_fmt='%s not in %s',
            )

        return v

    def not_in(self, v: T, c: ta.Container[T], msg: CheckMessage = None) -> T:
        if v in c:
            self._raise(
                ValueError,
                'Must not be in',
                msg,
                Checks._ArgsKwargs(v, c),
                render_fmt='%s in %s',
            )

        return v

    def empty(self, v: SizedT, msg: CheckMessage = None) -> SizedT:
        if len(v) != 0:
            self._raise(
                ValueError,
                'Must be empty',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )

        return v

    def iterempty(self, v: ta.Iterable[T], msg: CheckMessage = None) -> ta.Iterable[T]:
        it = iter(v)
        try:
            next(it)
        except StopIteration:
            pass
        else:
            self._raise(
                ValueError,
                'Must be empty',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )

        return v

    def not_empty(self, v: SizedT, msg: CheckMessage = None) -> SizedT:
        if len(v) == 0:
            self._raise(
                ValueError,
                'Must not be empty',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )

        return v

    def unique(self, it: ta.Iterable[T], msg: CheckMessage = None) -> ta.Iterable[T]:
        dupes = [e for e, c in collections.Counter(it).items() if c > 1]
        if dupes:
            self._raise(
                ValueError,
                'Must be unique',
                msg,
                Checks._ArgsKwargs(it, dupes),
            )

        return it

    def single(self, obj: ta.Iterable[T], msg: CheckMessage = None) -> T:
        try:
            [value] = obj
        except ValueError:
            self._raise(
                ValueError,
                'Must be single',
                msg,
                Checks._ArgsKwargs(obj),
                render_fmt='%s',
            )

        return value

    def opt_single(self, obj: ta.Iterable[T], msg: CheckMessage = None) -> ta.Optional[T]:
        it = iter(obj)
        try:
            value = next(it)
        except StopIteration:
            return None

        try:
            next(it)
        except StopIteration:
            return value  # noqa

        self._raise(
            ValueError,
            'Must be empty or single',
            msg,
            Checks._ArgsKwargs(obj),
            render_fmt='%s',
        )

        raise RuntimeError  # noqa

    #

    def none(self, v: ta.Any, msg: CheckMessage = None) -> None:
        if v is not None:
            self._raise(
                ValueError,
                'Must be None',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )

    def not_none(self, v: ta.Optional[T], msg: CheckMessage = None) -> T:
        if v is None:
            self._raise(
                ValueError,
                'Must not be None',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )

        return v

    #

    def equal(self, v: T, o: ta.Any, msg: CheckMessage = None) -> T:
        if o != v:
            self._raise(
                ValueError,
                'Must be equal',
                msg,
                Checks._ArgsKwargs(v, o),
                render_fmt='%s != %s',
            )

        return v

    def not_equal(self, v: T, o: ta.Any, msg: CheckMessage = None) -> T:
        if o == v:
            self._raise(
                ValueError,
                'Must not be equal',
                msg,
                Checks._ArgsKwargs(v, o),
                render_fmt='%s == %s',
            )

        return v

    def is_(self, v: T, o: ta.Any, msg: CheckMessage = None) -> T:
        if o is not v:
            self._raise(
                ValueError,
                'Must be the same',
                msg,
                Checks._ArgsKwargs(v, o),
                render_fmt='%s is not %s',
            )

        return v

    def is_not(self, v: T, o: ta.Any, msg: CheckMessage = None) -> T:
        if o is v:
            self._raise(
                ValueError,
                'Must not be the same',
                msg,
                Checks._ArgsKwargs(v, o),
                render_fmt='%s is %s',
            )

        return v

    def callable(self, v: T, msg: CheckMessage = None) -> T:  # noqa
        if not callable(v):
            self._raise(
                TypeError,
                'Must be callable',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )

        return v

    def non_empty_str(self, v: ta.Optional[str], msg: CheckMessage = None) -> str:
        if not isinstance(v, str) or not v:
            self._raise(
                ValueError,
                'Must be non-empty str',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )

        return v

    def replacing(self, expected: ta.Any, old: ta.Any, new: T, msg: CheckMessage = None) -> T:
        if old != expected:
            self._raise(
                ValueError,
                'Must be replacing',
                msg,
                Checks._ArgsKwargs(expected, old, new),
                render_fmt='%s -> %s -> %s',
            )

        return new

    def replacing_none(self, old: ta.Any, new: T, msg: CheckMessage = None) -> T:
        if old is not None:
            self._raise(
                ValueError,
                'Must be replacing None',
                msg,
                Checks._ArgsKwargs(old, new),
                render_fmt='%s -> %s',
            )

        return new

    #

    def arg(self, v: bool, msg: CheckMessage = None) -> None:
        if not v:
            self._raise(
                RuntimeError,
                'Argument condition not met',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )

    def state(self, v: bool, msg: CheckMessage = None) -> None:
        if not v:
            self._raise(
                RuntimeError,
                'State condition not met',
                msg,
                Checks._ArgsKwargs(v),
                render_fmt='%s',
            )


check = Checks()



LOREM = (
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore '
    'magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo '
    'consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. '
    'Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.'
)



# ruff: noqa: UP006 UP007 UP045
"""
https://docs.python.org/3/library/unittest.html#command-line-interface
~ https://github.com/python/cpython/tree/f66c75f11d3aeeb614600251fd5d3fe1a34b5ff1/Lib/unittest
"""
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001-2024 Python Software Foundation; All Rights
# Reserved" are retained in Python alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import dataclasses as dc
import types
import typing as ta
import unittest

from ...lite.abstract import Abstract
from .types import UnittestTest


##


class UnittestTargetLoader:
    def __init__(
            self,
            *,
            test_name_patterns: ta.Optional[ta.Sequence[str]] = None,
            module: ta.Union[str, types.ModuleType, None] = None,
            loader: ta.Optional[unittest.loader.TestLoader] = None,
    ) -> None:
        super().__init__()

        self._test_name_patterns = test_name_patterns
        self._module = module
        self._loader = loader

    #

    class Target(Abstract):
        pass

    class ModuleTarget(Target):
        pass

    @dc.dataclass(frozen=True)
    class NamesTarget(Target):
        test_names: ta.Optional[ta.Sequence[str]] = None

    @dc.dataclass(frozen=True)
    class DiscoveryTarget(Target):
        start: ta.Optional[str] = None
        pattern: ta.Optional[str] = None
        top: ta.Optional[str] = None

    def load(self, target: Target) -> UnittestTest:
        loader = self._loader
        if loader is None:
            loader = unittest.loader.TestLoader()

        if self._test_name_patterns:
            loader.testNamePatterns = self._test_name_patterns  # type: ignore[assignment]

        if isinstance(target, UnittestTargetLoader.DiscoveryTarget):
            return ta.cast(UnittestTest, loader.discover(
                target.start,  # type: ignore[arg-type]
                target.pattern,  # type: ignore[arg-type]
                target.top,
            ))

        module: ta.Any = self._module
        if isinstance(module, str):
            module = __import__(module)
            for part in module.split('.')[1:]:
                module = getattr(module, part)

        if isinstance(target, UnittestTargetLoader.ModuleTarget):
            return ta.cast(UnittestTest, loader.loadTestsFromModule(module))

        elif isinstance(target, UnittestTargetLoader.NamesTarget):
            return ta.cast(UnittestTest, loader.loadTestsFromNames(
                target.test_names,  # type: ignore[arg-type]
                module,
            ))

        else:
            raise TypeError(target)



import dataclasses as dc
import typing as ta

from .... import lang
from .base import Plan


##


@dc.dataclass(frozen=True)
class Plans:
    tup: tuple[Plan, ...]

    def __iter__(self) -> ta.Iterator[Plan]:
        return iter(self.tup)

    @lang.cached_function(no_wrapper_update=True)
    def repr(self) -> str:
        return repr(self)



import dataclasses as dc
import string
import typing as ta

from ...lite.check import check


##


@dc.dataclass()
class Cell:
    """Represents a single character cell in the terminal with attributes."""

    char: str = ' '

    fg: str = 'default'
    bg: str = 'default'

    bold: bool = False
    underline: bool = False
    reverse: bool = False

    def __repr__(self):
        return f'Cell({self.char!r}, bold={self.bold}, underline={self.underline})'


class Vt100Terminal:
    def __init__(
            self,
            rows: int = 24,
            cols: int = 80,
    ) -> None:
        super().__init__()

        self._rows = rows
        self._cols = cols

        # 2D array of Cell objects
        self._screen = [
            [Cell() for _ in range(cols)]
            for _ in range(rows)
        ]

        # Current cursor position (row, col), 0-based internally
        self._cursor_row = 0
        self._cursor_col = 0

        # Current text attributes
        self._current_fg = 'default'
        self._current_bg = 'default'
        self._current_bold = False
        self._current_underline = False
        self._current_reverse = False

        # Parser state
        self._state: ta.Literal['normal', 'esc', 'csi'] = 'normal'
        self._escape_buffer: list[str] = []

    def parse_byte(self, byte: int | str) -> None:
        """Parse a single byte of input (as an integer or a single-character string)."""

        if isinstance(byte, int):
            byte = chr(byte)

        if self._state == 'normal':
            if byte == '\x1b':
                # Start of escape sequence
                self._state = 'esc'
                self._escape_buffer = [byte]
            elif byte == '\r':
                # Carriage return
                self._cursor_col = 0
            elif byte == '\n':
                # Line feed
                self._cursor_row = min(self._cursor_row + 1, self._rows - 1)
            elif byte == '\b':
                # Backspace
                self._cursor_col = max(self._cursor_col - 1, 0)
            elif byte in string.printable and byte not in ['\x0b', '\x0c']:
                # Printable ASCII (excluding form feeds, vertical tabs, etc.)
                self._put_char(byte)
            else:
                # Ignore other control characters
                pass

        elif self._state == 'esc':
            self._escape_buffer.append(byte)
            if byte == '[':
                # Move to CSI state (Control Sequence Introducer)
                self._state = 'csi'
            # Some escape codes like ESCc, ESC7, etc. are possible but we'll ignore or handle them in a minimal way. If
            # no further instructions, revert to normal
            elif len(self._escape_buffer) == 2:
                # We only support ESC + [ in this example, so revert
                self._state = 'normal'

        elif self._state == 'csi':
            self._escape_buffer.append(byte)
            # Check if this byte ends the sequence (typical final bytes are A-Z, @, etc.)
            if byte.isalpha() or byte in '@`~':
                # We have a complete CSI sequence: parse it
                self._parse_csi(''.join(self._escape_buffer))
                # Reset state
                self._state = 'normal'
                self._escape_buffer = []

    def _put_char(self, ch: str) -> None:
        """Write a character at the current cursor position, and advance the cursor."""

        if 0 <= self._cursor_row < self._rows and 0 <= self._cursor_col < self._cols:
            cell = self._screen[self._cursor_row][self._cursor_col]
            cell.char = ch
            cell.fg = self._current_fg
            cell.bg = self._current_bg
            cell.bold = self._current_bold
            cell.underline = self._current_underline
            cell.reverse = self._current_reverse

        self._cursor_col += 1
        if self._cursor_col >= self._cols:
            self._cursor_col = 0
            self._cursor_row = min(self._cursor_row + 1, self._rows - 1)

    def _parse_csi(self, seq: str) -> None:
        """
        Parse a CSI (Control Sequence Introducer) escape sequence. Typically looks like: ESC [ parameters letter For
        example: ESC [ 2 J, ESC [ 10 ; 20 H, etc.
        """

        # seq includes the initial ESC[. e.g. "\x1b[10;20H"
        # We'll strip ESC[ and parse what remains.
        check.state(seq.startswith('\x1b['))
        body = seq[2:]  # everything after ESC[

        # Find final character that determines the command (e.g. 'A', 'B', 'H', 'J', 'K', 'm', etc.)
        final = body[-1]
        params = body[:-1]

        # Split params by semicolons, handle empty as '0'
        if params.strip() == '':
            numbers = [0]
        else:
            numbers = [int(x) if x.isdigit() else 0 for x in params.split(';')]

        if final in 'ABCD':
            # Cursor movement
            n = numbers[0] if numbers else 1
            if final == 'A':
                # Up
                self._cursor_row = max(self._cursor_row - n, 0)
            elif final == 'B':
                # Down
                self._cursor_row = min(self._cursor_row + n, self._rows - 1)
            elif final == 'C':
                # Right
                self._cursor_col = min(self._cursor_col + n, self._cols - 1)
            elif final == 'D':
                # Left
                self._cursor_col = max(self._cursor_col - n, 0)

        elif final in 'Hf':
            # Cursor position
            # CSI row;colH or CSI row;colf (1-based coords)
            row = numbers[0] if len(numbers) > 0 else 1
            col = numbers[1] if len(numbers) > 1 else 1
            self._cursor_row = min(max(row - 1, 0), self._rows - 1)
            self._cursor_col = min(max(col - 1, 0), self._cols - 1)

        elif final == 'J':
            # Erase display
            # n=0 -> clear from cursor down,
            # n=1 -> clear from cursor up,
            # n=2 -> clear entire screen
            n = numbers[0] if numbers else 0
            if n == 2:
                self._clear_screen()
            elif n == 0:
                self._clear_down()
            elif n == 1:
                self._clear_up()
            # else: unsupported J mode, ignore

        elif final == 'K':
            # Erase line
            # n=0 -> clear from cursor right
            # n=1 -> clear from cursor left
            # n=2 -> clear entire line
            n = numbers[0] if numbers else 0
            if n == 2:
                self._clear_line(self._cursor_row)
            elif n == 0:
                self._clear_right(self._cursor_row, self._cursor_col)
            elif n == 1:
                self._clear_left(self._cursor_row, self._cursor_col)
            # else: ignore

        elif final == 'm':
            # SGR - Select Graphic Rendition
            # We handle a subset: 0 (reset), 1 (bold), 4 (underline), 7 (reverse)
            # Colors could be extended, but here we keep it minimal
            for code in numbers:
                if code == 0:
                    self._current_fg = 'default'
                    self._current_bg = 'default'
                    self._current_bold = False
                    self._current_underline = False
                    self._current_reverse = False
                elif code == 1:
                    self._current_bold = True
                elif code == 4:
                    self._current_underline = True
                elif code == 7:
                    self._current_reverse = True
                # You can add more codes for color, etc.
                else:
                    # Unsupported SGR code - ignore gracefully
                    pass

        else:
            # Unsupported final - ignore gracefully
            pass

    # hods for Erase Operations

    def _clear_screen(self) -> None:
        for r in range(self._rows):
            for c in range(self._cols):
                self._screen[r][c] = Cell()

    def _clear_down(self) -> None:
        """Clear from cursor to the end of the screen."""

        # Clear current line from cursor forward
        self._clear_right(self._cursor_row, self._cursor_col)

        # Clear all lines below cursor
        for r in range(self._cursor_row + 1, self._rows):
            self._clear_line(r)

    def _clear_up(self) -> None:
        """Clear from the start of the screen up to the cursor."""

        # Clear current line from start to cursor
        self._clear_left(self._cursor_row, self._cursor_col)

        # Clear all lines above cursor
        for r in range(self._cursor_row):
            self._clear_line(r)

    def _clear_line(self, row: int) -> None:
        for c in range(self._cols):
            self._screen[row][c] = Cell()

    def _clear_right(self, row: int, col: int) -> None:
        for c in range(col, self._cols):
            self._screen[row][c] = Cell()

    def _clear_left(self, row: int, col: int) -> None:
        for c in range(col + 1):
            self._screen[row][c] = Cell()

    # Debug/Utility Methods

    def get_screen_as_strings(self) -> list[str]:
        """Return a list of strings representing each row (ignoring attributes). Useful for debugging/testing."""

        return [''.join(cell.char for cell in row) for row in self._screen]



from .consts import (  # noqa
    DEFAULT_BUFFER_SIZE,
)

from .direct import (  # noqa
    DirectCoro,

    BytesDirectCoro,
    StrDirectCoro,
)

from .readers import (  # noqa
    ReaderCoro,
    BytesReaderCoro,
    StrReaderCoro,

    ExactReaderCoro,
    BytesExactReaderCoro,
    StrExactReaderCoro,

    CoroReader,

    PrependableCoroReader,
    PrependableBytesCoroReader,
    PrependableStrCoroReader,
    prependable_bytes_coro_reader,
    prependable_str_coro_reader,

    BufferedCoroReader,
    BufferedBytesCoroReader,
    BufferedStrCoroReader,
    buffered_bytes_coro_reader,
    buffered_str_coro_reader,
)

from .stepped import (  # noqa
    SteppedCoro,
    BytesSteppedCoro,
    StrSteppedCoro,
    BytesToStrSteppedCoro,
    StrToBytesSteppedCoro,

    SteppedReaderCoro,
    BytesSteppedReaderCoro,
    StrSteppedReaderCoro,

    flatmap_stepped_coro,

    joined_bytes_stepped_coro,
    joined_str_stepped_coro,

    read_into_bytes_stepped_coro,
    read_into_str_stepped_coro,

    buffer_bytes_stepped_reader_coro,

    iterable_bytes_stepped_coro,
)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import asyncio
import typing as ta


##


class AsyncioBytesChannelTransport(asyncio.Transport):
    def __init__(self, reader: asyncio.StreamReader) -> None:
        super().__init__()

        self.reader = reader
        self.closed: asyncio.Future = asyncio.Future()

    # @ta.override
    def write(self, data: bytes) -> None:
        self.reader.feed_data(data)

    # @ta.override
    def close(self) -> None:
        self.reader.feed_eof()
        if not self.closed.done():
            self.closed.set_result(True)

    # @ta.override
    def is_closing(self) -> bool:
        return self.closed.done()


def asyncio_create_bytes_channel(
        loop: ta.Any = None,
) -> ta.Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
    if loop is None:
        loop = asyncio.get_running_loop()

    reader = asyncio.StreamReader()
    protocol = asyncio.StreamReaderProtocol(reader)
    transport = AsyncioBytesChannelTransport(reader)
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)

    return reader, writer



import dataclasses as dc
import functools
import typing as ta

from ... import lang
from ..coro import BytesSteppedCoro
from ..coro import BytesSteppedReaderCoro
from .adapters import CompressorObjectIncrementalAdapter
from .adapters import DecompressorObjectIncrementalAdapter
from .base import Compression
from .base import IncrementalCompression
from .codecs import make_compression_codec
from .codecs import make_compression_lazy_loaded_codec


if ta.TYPE_CHECKING:
    import bz2
else:
    bz2 = lang.proxy_import('bz2')


##


@dc.dataclass(frozen=True, kw_only=True)
class Bz2Compression(Compression, IncrementalCompression):
    level: int = 9

    def compress(self, d: bytes) -> bytes:
        return bz2.compress(
            d,
            self.level,
        )

    def decompress(self, d: bytes) -> bytes:
        return bz2.decompress(
            d,
        )

    def compress_incremental(self) -> BytesSteppedCoro[None]:
        return lang.nextgen(CompressorObjectIncrementalAdapter(
            functools.partial(
                bz2.BZ2Compressor,  # type: ignore
                self.level,
            ),
        )())

    def decompress_incremental(self) -> BytesSteppedReaderCoro[None]:
        return DecompressorObjectIncrementalAdapter(
            bz2.BZ2Decompressor,  # type: ignore
            trailing_error=OSError,
        )()


##


BZ2_CODEC = make_compression_codec('bz2', Bz2Compression)

# @omlish-manifest
_BZ2_LAZY_CODEC = make_compression_lazy_loaded_codec(__name__, 'BZ2_CODEC', BZ2_CODEC)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import subprocess
import sys
import typing as ta

from ..lite.abstract import Abstract
from ..lite.check import check
from ..lite.timeouts import TimeoutLike
from .base import BaseSubprocesses
from .run import SubprocessRun
from .run import SubprocessRunOutput


##


class AbstractAsyncSubprocesses(BaseSubprocesses, Abstract):
    @abc.abstractmethod
    def run_(self, run: SubprocessRun) -> ta.Awaitable[SubprocessRunOutput]:
        raise NotImplementedError

    def run(
            self,
            *cmd: str,
            input: ta.Any = None,  # noqa
            timeout: TimeoutLike = None,
            check: bool = False,
            capture_output: ta.Optional[bool] = None,
            **kwargs: ta.Any,
    ) -> ta.Awaitable[SubprocessRunOutput]:
        return self.run_(SubprocessRun(
            cmd=cmd,
            input=input,
            timeout=timeout,
            check=check,
            capture_output=capture_output,
            kwargs=kwargs,
        ))

    #

    async def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> None:
        await self.run(
            *cmd,
            stdout=stdout,
            check=True,
            **kwargs,
        )

    async def check_output(
            self,
            *cmd: str,
            stdout: ta.Any = subprocess.PIPE,
            **kwargs: ta.Any,
    ) -> bytes:
        return check.not_none((await self.run(
            *cmd,
            stdout=stdout,
            check=True,
            **kwargs,
        )).stdout)

    async def check_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> str:
        return (await self.check_output(
            *cmd,
            **kwargs,
        )).decode().strip()

    #

    async def try_call(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bool:
        if isinstance(await self.async_try_fn(self.check_call, *cmd, **kwargs), Exception):
            return False
        else:
            return True

    async def try_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[bytes]:
        if isinstance(ret := await self.async_try_fn(self.check_output, *cmd, **kwargs), Exception):
            return None
        else:
            return ret

    async def try_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[str]:
        if (ret := await self.try_output(*cmd, **kwargs)) is None:
            return None
        else:
            return ret.decode().strip()



"""
A generator powered, configurable, mostly fully streaming JSON parser.

Regarding the 'streamyness' of the subsystems:
 - Lexing only buffers for string and number literals.
 - Parsing maintains only a stack that scales by nesting depth.
 - Building values will obviously hold everything under the topmost object it's building until it's finished.

It's reasonably optimized, but performance is not a primary or even secondary goal: its goal is flexibility. If speed
matters use a native library.
"""


from .building import (  # noqa
    JsonValueBuilder,
)

from .errors import (  # noqa
    JsonStreamError,
)


from .lexing import (  # noqa
    IdentTokenKind,
    ValueTokenKind,
    VALUE_TOKEN_KINDS,
    ControlTokenKind,
    SpaceTokenKind,
    TokenKind,

    ScalarValue,
    SCALAR_VALUE_TYPES,

    Position,

    Token,

    CONTROL_TOKENS,
    CONST_IDENT_VALUES,

    JsonStreamLexError,
    JsonStreamLexer,
)

from .parsing import (  # noqa
    BeginObject,
    Key,
    EndObject,
    BeginArray,
    EndArray,

    Event,
    Events,

    yield_parser_events,

    JsonStreamParseError,
    JsonStreamObject,
    JsonStreamParser,
)

from .rendering import (  # noqa
    StreamJsonRenderer,
)

from .utils import (  # noqa
    JsonStreamValueParser,

    stream_parse_values,
    stream_parse_one_value,
    stream_parse_exactly_one_value,
)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import sys
import typing as ta

from ..lite.maysync import make_maysync
from .asyncs import AbstractAsyncSubprocesses
from .run import SubprocessRun
from .run import SubprocessRunOutput
from .sync import AbstractSubprocesses


##


class MaysyncSubprocesses(AbstractAsyncSubprocesses):
    def __init__(
            self,
            subprocesses: AbstractSubprocesses,
            async_subprocesses: AbstractAsyncSubprocesses,
    ) -> None:
        super().__init__()

        self._subprocesses = subprocesses
        self._async_subprocesses = async_subprocesses

    #

    def run_(self, run: SubprocessRun) -> ta.Awaitable[SubprocessRunOutput]:
        return make_maysync(
            self._subprocesses.run,
            self._async_subprocesses.run,
        )(run)

    #

    async def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> None:
        return await make_maysync(
            self._subprocesses.check_call,
            self._async_subprocesses.check_call,
        )(*cmd, stdout=stdout, **kwargs)

    async def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bytes:
        return await make_maysync(
            self._subprocesses.check_output,
            self._async_subprocesses.check_output,
        )(*cmd, **kwargs)



import typing as ta

from ... import cached
from ... import dataclasses as dc


##


class RouteKey(ta.NamedTuple):
    rule: str
    method: str


@dc.dataclass(frozen=True, kw_only=True)
class Route:
    rule: str
    endpoint: str
    methods: frozenset[str]

    @cached.property
    def keys(self) -> frozenset[RouteKey]:
        return frozenset(RouteKey(self.rule, m) for m in self.methods)



"""
TODO:
 - try_weak .. allow non-weakrefable keys to just fall through and not cache
"""
import enum
import functools
import typing as ta
import weakref

from .impl import new_cache
from .types import Cache


C = ta.TypeVar('C', bound=ta.Callable)
CC: ta.TypeAlias = ta.Callable[[C], C]


##


class Scope(enum.Enum):
    INSTANCE = 'INSTANCE'
    CLASS = 'CLASS'
    STATIC = 'STATIC'


class _HashedSeq(list):
    __slots__ = ['hash_value']

    def __init__(
            self,
            tup: tuple,
            hasher: ta.Callable[[ta.Any], int] = hash,
    ) -> None:
        super().__init__()

        self[:] = tup
        self.hash_value = hasher(tup)

    def __hash__(self) -> int:  # type: ignore
        return self.hash_value


def _make_key(
        args: tuple,
        kwargs: dict[str, ta.Any],
        typed: bool,
        kwd_mark=(object(),),
        fasttypes=frozenset([int, str, frozenset, type(None)]),
        tuple=tuple,  # noqa
        type=type,  # noqa
        len=len,  # noqa
) -> ta.Any:
    key = args
    if kwargs:
        key += kwd_mark
        for item in kwargs.items():
            key += item
    if typed:
        key += tuple(type(v) for v in args)
        if kwargs:
            key += tuple(type(v) for v in kwargs.values())
    elif len(key) == 1 and type(key[0]) in fasttypes:
        return key[0]
    return _HashedSeq(key)


class Ignore:
    def __init__(self, value: ta.Any) -> None:
        super().__init__()

        self._value = value


def ignore(value: ta.Any) -> ta.Any:
    return Ignore(value)


class _CacheDescriptor:
    def __init__(
            self,
            fn: ta.Callable,
            scope: Scope,
            typed: bool,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__()

        self._fn = fn
        functools.update_wrapper(self, fn)
        self._scope = scope
        self._typed = typed
        self._kwargs = kwargs
        self.__static: Cache | None = None
        self.__static_built: ta.Callable | None = None
        self._by_class: ta.MutableMapping[ta.Type, Cache] | None = weakref.WeakKeyDictionary() if scope == Scope.CLASS else None  # noqa
        self._name = None
        self._unary = kwargs.get('identity_keys', False) or kwargs.get('weak_keys', False)

    def __set_name__(self, owner, name):
        if self._name is None:
            self._name = name

    @property
    def _static(self) -> Cache:
        if self.__static is None:
            self.__static = new_cache(**self._kwargs)
        return self.__static

    def _build(self, fn: ta.Callable, cache: Cache):
        def miss(key, result):
            if isinstance(result, Ignore):
                return result._value  # noqa
            else:
                cache[key] = result
                return result

        if self._unary:
            @functools.wraps(fn)
            def inner(key):
                try:
                    return cache[key]
                except KeyError:
                    pass
                return miss(key, fn(key))

        else:
            @functools.wraps(fn)
            def inner(*args, **kwargs):
                key = _make_key(args, kwargs, self._typed)
                try:
                    return cache[key]
                except KeyError:
                    pass
                return miss(key, fn(*args, **kwargs))

        return inner

    def __get__(self, instance, owner=None):
        if self._scope == Scope.STATIC:
            cache = self._static

        elif self._scope == Scope.CLASS:
            if owner is None:
                raise TypeError
            try:
                cache = self._by_class[owner]  # type: ignore
            except KeyError:
                cache = self._by_class[owner] = new_cache(**self._kwargs)  # type: ignore

        elif self._scope == Scope.INSTANCE:
            if instance is not None:
                cache = new_cache()
            else:
                @functools.wraps(self._fn)
                def trampoline(this, *args, **kwargs):
                    return self.__get__(this, owner)(*args, **kwargs)
                return trampoline

        else:
            raise TypeError

        fn = self._build(self._fn.__get__(instance, owner), cache)  # noqa

        if self._scope == Scope.CLASS:
            setattr(owner, self._name, fn)  # type: ignore
        elif self._scope == Scope.INSTANCE:
            setattr(instance, self._name, fn)  # type: ignore

        return fn

    def __call__(self, *args, **kwargs):
        if self.__static_built is None:
            self.__static_built = self._build(self._fn, self._static)
        return self.__static_built(*args, **kwargs)


def cache(
        scope: Scope | str = Scope.INSTANCE,
        typed: bool = False,
        **kwargs,
) -> CC:
    if not isinstance(scope, Scope):
        scope = getattr(Scope, scope.upper())  # noqa

    def inner(fn):
        return _CacheDescriptor(fn, scope, typed, **kwargs)  # type: ignore

    return inner



from .. import lang as _lang


with _lang.auto_proxy_init(
        globals(),
        update_exports=True,
        # eager=True,
):
    ##

    from .inspect import (  # noqa
        has_annotations,

        get_annotations,

        get_filtered_type_hints,
    )

    from .ops import (  # noqa
        strip_objs,
        strip_annotations,

        types_equivalent,

        get_underlying,
        get_concrete_type,

        to_annotation,
    )

    from .subst import (  # noqa
        get_type_var_replacements,
        replace_type_vars,

        GenericSubstitution,
        DEFAULT_GENERIC_SUBSTITUTION,
        ALIAS_UPDATING_GENERIC_SUBSTITUTION,

        generic_mro,
        get_generic_bases,
    )

    from .types import (  # noqa
        is_simple_generic_alias_type,
        get_params,
        is_union_type,
        get_orig_bases,
        get_orig_class,
        get_newtype_supertype,
        get_type_var_bound,

        TypeInfo,
        Type,
        TYPES,

        Union,

        GenericLike,
        Generic,
        Protocol,

        NewType,

        Annotated,

        Literal,

        Any,
        ANY,

        ReflectTypeError,

        Reflector,
        DEFAULT_REFLECTOR,
        is_type,
        type_,
    )



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import dataclasses as dc
import http.server
import logging
import typing as ta

from ..lite.abstract import Abstract
from ..logs.protocols import LoggerLike
from ..sockets.addresses import SocketAddress
from .parsing import HttpHeaders


HttpHandler = ta.Callable[['HttpHandlerRequest'], 'HttpHandlerResponse']  # ta.TypeAlias
HttpHandlerResponseData = ta.Union[bytes, 'HttpHandlerResponseStreamedData']  # ta.TypeAlias  # noqa


##


@dc.dataclass(frozen=True)
class HttpHandlerRequest:
    client_address: SocketAddress
    method: str
    path: str
    headers: HttpHeaders
    data: ta.Optional[bytes]


@dc.dataclass(frozen=True)
class HttpHandlerResponse:
    status: ta.Union[http.HTTPStatus, int]

    headers: ta.Optional[ta.Mapping[str, str]] = None
    data: ta.Optional[HttpHandlerResponseData] = None
    close_connection: ta.Optional[bool] = None

    def close(self) -> None:
        if isinstance(d := self.data, HttpHandlerResponseStreamedData):
            d.close()


@dc.dataclass(frozen=True)
class HttpHandlerResponseStreamedData:
    iter: ta.Iterable[bytes]
    length: ta.Optional[int] = None

    def close(self) -> None:
        if hasattr(d := self.iter, 'close'):
            d.close()  # noqa


class HttpHandlerError(Exception):
    pass


class UnsupportedMethodHttpHandlerError(Exception):
    pass


class HttpHandler_(Abstract):  # noqa
    @abc.abstractmethod
    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class LoggingHttpHandler(HttpHandler_):
    handler: HttpHandler
    log: LoggerLike
    level: int = logging.DEBUG

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        self.log.log(self.level, '%r', req)
        resp = self.handler(req)
        self.log.log(self.level, '%r', resp)
        return resp


@dc.dataclass(frozen=True)
class ExceptionLoggingHttpHandler(HttpHandler_):
    handler: HttpHandler
    log: LoggerLike
    message: ta.Union[str, ta.Callable[[HttpHandlerRequest, BaseException], str]] = 'Error in http handler'

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        try:
            return self.handler(req)
        except Exception as e:  # noqa
            if callable(msg := self.message):
                msg = msg(req, e)
            self.log.exception(msg)
            raise


##


@dc.dataclass(frozen=True)
class BytesResponseHttpHandler(HttpHandler_):
    data: bytes

    status: ta.Union[http.HTTPStatus, int] = 200
    content_type: ta.Optional[str] = 'application/octet-stream'
    headers: ta.Optional[ta.Mapping[str, str]] = None
    close_connection: bool = True

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        return HttpHandlerResponse(
            status=self.status,
            headers={
                **({'Content-Type': self.content_type} if self.content_type else {}),
                'Content-Length': str(len(self.data)),
                **(self.headers or {}),
            },
            data=self.data,
            close_connection=self.close_connection,
        )


@dc.dataclass(frozen=True)
class StringResponseHttpHandler(HttpHandler_):
    data: str

    status: ta.Union[http.HTTPStatus, int] = 200
    content_type: ta.Optional[str] = 'text/plain; charset=utf-8'
    headers: ta.Optional[ta.Mapping[str, str]] = None
    close_connection: bool = True

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        data = self.data.encode('utf-8')
        return HttpHandlerResponse(
            status=self.status,
            headers={
                **({'Content-Type': self.content_type} if self.content_type else {}),
                'Content-Length': str(len(data)),
                **(self.headers or {}),
            },
            data=data,
            close_connection=self.close_connection,
        )



import logging


##


NOISY_LOGGER_LEVELS: dict[str, int] = {
    'boto3.resources.action': logging.WARNING,
    'datadog.dogstatsd': logging.WARNING,
    'elasticsearch': logging.WARNING,
    'httpcore': logging.INFO,
    'httpx': logging.WARNING,
    'kazoo.client': logging.WARNING,
    'markdown_it': logging.INFO,
    'requests.packages.urllib3.connectionpool': logging.WARNING,
}


def silence_noisy_loggers() -> None:
    for name, level in NOISY_LOGGER_LEVELS.items():
        logging.getLogger(name).setLevel(level)
